/* safePrint.c
   Scott F. Kaplan -- sfkaplan@cs.utexas.edu
   February 1997
   This module is used to print values to a string.  It guarantees that
   only stack space is used. */


#include "safePrint.h"

#define FALSE 0
#define TRUE 1

void safePrintPageNumber (char* bufferString,
			  int dirtyStatus,
			  unsigned long printValue) {
  int hexPosition;
  int bufferIndex;
  unsigned long mask;
  unsigned long maskedResult;
  unsigned long shiftedResult;

  /* Initialize the buffer index to zero. */
  bufferIndex = 0;

  /* What is the dirty status of this page?  (And thus, what prefix should
     go in front of the page number output?) */
  switch (dirtyStatus) {

  case PAGE_STATUS_NONE:
    break;

  case PAGE_STATUS_READ:
    bufferString[bufferIndex++] = 'R';
    bufferString[bufferIndex++] = ' ';
    break;

  case PAGE_STATUS_WRITE:
    bufferString[bufferIndex++] = 'W';
    bufferString[bufferIndex++] = ' ';
    break;

  default:
    exit(-1);
    
  }
  
  /* Initialize the bitmask to mask out all but the most significant
     hexidecimial character. */
  mask = 0x0000F000;

  /* Loop through each hex character, from most significant to least. */
  for (hexPosition = 3; hexPosition >= 0; hexPosition--) {

    /* Apply the mask to the address to isolate the bits that
       represent the hex character that we want to turn into a
       character. */
    maskedResult = printValue & mask;

    /* Shift the isolated bits down to the least significant hex
       position. */
    shiftedResult = maskedResult >> (hexPosition * 4);

    /* Is the shifted value between zero and nine? */
    if ((shiftedResult >= 0) && (shiftedResult <= 9)) {
      
      /* It is, so add the proper ASCII offset to turn it into
	 a character. */
      bufferString[bufferIndex] = shiftedResult + 48;
      
    } else if ((shiftedResult >= 10) && (shiftedResult <= 15)) {
      
      /* It's between A and F, so add the proper ASCII offset
	 to turn that hexidecimal value into a character. */
      bufferString[bufferIndex] = shiftedResult + 87;
      
    } else {
      
      exit(-1);
      
    }
      
    /* The current hex character has been turned into a character.
       Advance to the next position in the string. */
    bufferIndex++;
      
    /* Shift the mask for the next less significant hexidecimal
       character. */
    mask = mask >> 4;
    
  }

  bufferString[bufferIndex++] = '\n';
  bufferString[bufferIndex] = '\0';

}


void safePrintLong (char* bufferString, unsigned long printValue) {
  int hexPosition;
  int pastLeadingZeros;
  int bufferIndex;
  unsigned long mask;
  unsigned long maskedResult;
  unsigned long shiftedResult;

  /* Initialize the buffer index to zero. */
  bufferIndex = 0;

  /* Initialize the bitmask to mask out all but the most significant
     hexidecimial character. */
  mask = 0xF0000000;

  /* Loop through each hex character, from most significant to least. */
  for (hexPosition = 7, pastLeadingZeros = FALSE;
       hexPosition >= 0;
       hexPosition--) {

    /* Apply the mask to the address to isolate the bits that
       represent the hex character that we want to turn into a
       character. */
    maskedResult = printValue & mask;

    /* Shift the isolated bits down to the least significant hex
       position. */
    shiftedResult = maskedResult >> (hexPosition * 4);

    /* Have we just reached the first meaningful digit? */
    if (((shiftedResult > 0) && (!pastLeadingZeros)) ||
	(hexPosition == 0)) {

      /* We have, so from now on, print zeros. */
      pastLeadingZeros = TRUE;

    }

    /* Are we past the leading zeros? */
    if (pastLeadingZeros) {    
      
      /* Is the shifted value between zero and nine? */
      if ((shiftedResult >= 0) && (shiftedResult <= 9)) {
	
	/* It is, so add the proper ASCII offset to turn it into
	   a character. */
	bufferString[bufferIndex] = shiftedResult + 48;
	
      } else if ((shiftedResult >= 10) && (shiftedResult <= 15)) {
	
	/* It's between A and F, so add the proper ASCII offset
	   to turn that hexidecimal value into a character. */
	bufferString[bufferIndex] = shiftedResult + 87;
	
      } else {
	
	exit(-1);
	
      }

      /* The current hex character has been turned into a character.
	 Advance to the next position in the string. */
      bufferIndex++;

    }
      
    /* Shift the mask for the next less significant hexidecimal
       character. */
    mask = mask >> 4;
    
  }

  /* Add a newline and null terminator to the string. */
  bufferString[bufferIndex++] = '\n';
  bufferString[bufferIndex] = '\0';

}



void safePrintLongLong (char* bufferString, unsigned long long printValue) {
  int hexPosition;
  int pastLeadingZeros;
  int bufferIndex;
  unsigned long long mask;
  unsigned long long maskedResult;
  unsigned long long shiftedResult;

  /* Initialize the buffer index to zero. */
  bufferIndex = 0;

  /* Initialize the bitmask to mask out all but the most significant
     hexidecimial character.  Set the lowest four bits and shift them
     up--gcc/g++ don't seem to handle 64-bit constants well. */
  mask = 0xF;
  mask = mask << 60;

  /* Loop through each hex character, from most significant to least. */
  for (hexPosition = 15, pastLeadingZeros = FALSE;
       hexPosition >= 0;
       hexPosition--) {

    /* Apply the mask to the address to isolate the bits that
       represent the hex character that we want to turn into a
       character. */
    maskedResult = printValue & mask;

    /* Shift the isolated bits down to the least significant hex
       position. */
    shiftedResult = maskedResult >> (hexPosition * 4);

    /* Have we just reached the first meaningful digit? */
    if ((shiftedResult > 0) && (!pastLeadingZeros)) {

      /* We have, so from now on, print zeros. */
      pastLeadingZeros = TRUE;

    }

    /* Are we past the leading zeros? */
    if (pastLeadingZeros) {

      /* Is the shifted value between zero and nine? */
      if ((shiftedResult >= 0) && (shiftedResult <= 9)) {
	
	/* It is, so add the proper ASCII offset to turn it into
	   a character. */
	bufferString[bufferIndex] = shiftedResult + 48;
	
      } else if ((shiftedResult >= 10) && (shiftedResult <= 15)) {
	
	/* It's between A and F, so add the proper ASCII offset
	   to turn that hexidecimal value into a character. */
	bufferString[bufferIndex] = shiftedResult + 87;
	
      } else {
	
	exit(-1);
	
      }

      /* The current hex character has been turned into a character.
	 Advance to the next position in the string. */
      bufferIndex++;
      
    }

    /* Shift the mask for the next less significant hexidecimal
       character. */
    mask = mask >> 4;
    
  }

  /* Add a newline and null terminator to the string. */
  bufferString[bufferIndex++] = '\n';
  bufferString[bufferIndex] = '\0';

}
