/* traceContext.h -- Scott F. Kaplan <sfkaplan@cs.utexas.edu>
   Last modified:  May 1997

   This header file is intended for all data that the tracer or
   experiment needs to persist throughout the trace.  Special types
   and structures that need to persist throughout the trace should be
   declared first.  Then, any such data, of primitive or user-defined
   types, should be declared within the "TraceContext" structure.
   Finally, the TraceContextContainer is defined, and an extern
   reference to it is given so that all modules that need such
   persistent data can access the TraceContext--there will be only
   one, and it will be memory mapped so that it is not traced.  The
   corresponding code file for this module will contain only the
   global data item for the trace context container. */



#if !defined (_traceContext_h)
#define _traceContext_h

#include <sys/types.h>
#include "sysparams.h"


/*
** Generic tracing options and parameters
*/
#define MAX_UNPROTECTED_SET_SIZE 256
#define DEFAULT_UNPROTECTED_SET_SIZE 8
#define DEFAULT_OUTFILE "vmtrace.out"
#define INTERNAL_WRITE_BUFFER_SIZE 256
#define FILE_NAME_LEN 64
#define MAPPED_FILE_NAME "vmtrace.map"


/* A structure that holds information for each page in the
   unprotected set. */
typedef struct {
  caddr_t pageAddress;
  unsigned short dirty;
  /* Added elements for dumping pages. */
  VMT_word fetchedPageImage [WORDS_PER_PAGE];
} PageInfoElement;

/* A structure that holds the information for each element
   in the unprotected set. */
typedef struct {
  PageInfoElement pages [MAX_UNPROTECTED_SET_SIZE];
  unsigned int size;
  unsigned int length;
  unsigned int head;
} UnprotectedSet;


/* A type to store the status of various options. */
typedef enum {
  DISABLED,
  ENABLED
} TraceStatus;


/* A type to represent the replacement policies that can be used on
   the set of unprotected pages. */
typedef enum {
  FIFO,
  RANDOM
} ReplacementType;


/* A structure that will hold the state of the tracer. */
typedef struct {
  TraceStatus dirtyStatus; /* The status of tracing page dirtiness. */
  TraceStatus staticStatus; /* The status of tracing the static data
			       segment. */
  TraceStatus exptStatus; /* Indicates whether an experiemtn is
			     attatched to this trace. */
  ReplacementType replacement;
  UnprotectedSet unprotectedSet; /* The set of unprotected pages. */

  char wbuf [INTERNAL_WRITE_BUFFER_SIZE]; /* tmp write buffer */
  int fd;             /* file descriptor for trace output */
  char filename [FILE_NAME_LEN];

  caddr_t abrk, ubrk; /* Actual and user breaks (ends of heap) */
  caddr_t start_of_heap; /* The address at which the heap begins. */

  int inHandler; /* A flag to track nested faults. */
  caddr_t dataSegmentBegin; /* The address at which the static data
			       begins. */
  caddr_t dataSegmentEnd; /* The address at which the static data
			     ends. */
  unsigned long randomValue; /* State information needed for the
				random number generator. */
  PageInfoElement evictedPage; /* Used for experiments. */
} TraceContext;


/* A kludge to avoid nested faults when grabbing the trace context
   pointer within the fault handler.  Allocate about a page on either
   side of the pointer to guarantee that the pointer to the trace
   context isn't shared with any other items in the static data
   segment. */
typedef struct {
  char a[PAGESIZE / sizeof(char)];
  TraceStatus status; /* The status of tracing module */
  TraceContext* tc;
  char b[((PAGESIZE - sizeof(TraceContext*) - sizeof(TraceStatus)) /
	  sizeof(char))];
} TraceContextContainer;

extern TraceContextContainer tcc;

#endif /* _traceContext_h */
