/* ---------------------------------------------------------------
  The tracing mechanisms, including the initialization of VMTrace, the
  setting of a signal handler, that signal handler (which records the
  trace), and the clean-up functions.

  Author: V. B. Balayoghan (vbb@cs.utexas.edu)
  Date: October 1994
  Modified by: Scott F. Kaplan <sfkaplan@cs.utexas.edu>
  Last modification: September 1998
   --------------------------------------------------------------- */



/* --------------------------------------------------------------- */
#ifdef __linuxWithOldLibc5__
#define __KERNEL__
#include <linux/signal.h>
#undef __KERNEL__
#endif /* __linuxWithOldLibc5__ */

#include <stdio.h>
#include <signal.h>
#include <memory.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <fcntl.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include "traceContext.h"
#include "safePrint.h"
#include "BSDrand.h"

/* Modify this file if you want to attach an experiment to VMTrace. */
#include "expt.h"

/* Wrapped system calls that need to be called directly
   (circumventing their wrappers). */
#include "wrap_write.h"
#include "wrap_sbrk.h"
/* --------------------------------------------------------------- */



/* --------------------------------------------------------------- */
extern caddr_t _etext;
extern caddr_t _end;
#if defined (__svr4__)
extern caddr_t _siguhandler;
#endif /* __svr4__ */
/* --------------------------------------------------------------- */



/* --------------------------------------------------------------- */
/*
** Print given message, error from errno, and then abort
*/
void Abort(char* msg)
{
  perror(msg);
  abort();
}
/* --------------------------------------------------------------- */

 

/* --------------------------------------------------------------- */
/* Initialize the unprotected set.  Simply initialize the page
   information for each entry in the set. */
void
initUnprotectedSet (UnprotectedSet* newSet, int size) {

  unsigned int counter;

  newSet->size = size;
  newSet->length = 0;
  newSet->head = 0;

  for (counter = 0; counter < newSet->size; counter++) {
    newSet->pages[counter].pageAddress = 0;
    newSet->pages[counter].dirty = FALSE;
  }
}
/* --------------------------------------------------------------- */



/* --------------------------------------------------------------- */
/* Given an unprotected set (addSet) and the address at which a
   faulted page begins (faultPageAddress), add that page to the set
   and mark it as clean.  If the set was full, select the oldest page
   and return its information. */
PageInfoElement*
addToUnprotectedSet (UnprotectedSet* addSet, caddr_t faultPageAddress) {

  /* Is the set full? */
  if (addSet->size == addSet->length) {

    /* It is, so evict the page at which head points, insert the new
       page, and update the head pointer. */
    tcc.tc->evictedPage = addSet->pages[addSet->head];
    addSet->pages[addSet->head].pageAddress = faultPageAddress;
    addSet->pages[addSet->head].dirty = FALSE;

    /* If an experiment is attatched, call the hooks into the experiment. */
    if (tcc.tc->exptStatus == ENABLED) {
      ExptFetchEvent (&(addSet->pages[addSet->head]));
      ExptEvictEvent (&(tcc.tc->evictedPage));
    }

    /* Which policy is being used for replacement? */
    if (tcc.tc->replacement == RANDOM) {
      unsigned long rand_max = RAND_MAX;

      /* Select a random element to be the head. */
      addSet->head = (unsigned long)(((double)addSet->size
				      * BSDrand())
				     / (double)rand_max);

    } else {

      /* Advance the head pointer to the next element, and wrap the
	 pointer if needed. */
      addSet->head = (addSet->head + 1) % addSet->size;

    }    
  } else {

    /* It is not full, so use an unused slot. */
    tcc.tc->evictedPage.pageAddress = (caddr_t)0;
    addSet->pages[addSet->length].pageAddress = faultPageAddress;
    addSet->pages[addSet->length].dirty = FALSE;
    
    /* If an experiment is attatched, call the fault hook. */
    if (tcc.tc->exptStatus == ENABLED) {
      ExptFetchEvent (&(addSet->pages[addSet->length]));
    }

    addSet->length++;

  }

  return &(tcc.tc->evictedPage);

}
/* --------------------------------------------------------------- */



/* --------------------------------------------------------------- */
/* Given an unprotected (searchSet) and the beginning address of a
   page for which to search, try to find that page within the set.  If
   it is found, return the index at which it resides in the set.  If
   not, return -1. */
int searchUnprotectedSet (UnprotectedSet* searchSet,
			  caddr_t searchPageAddress) {
  int index;
  int found;

  /* Loop through each slot used in the unprotected set until the
     correct index is found or we reach the end of the set. */
  index = 0;
  found = FALSE;
  while ((index < (signed int)searchSet->length) && (!found)) {

    /* Is the page at this index in the set? */
    if (searchSet->pages[index].pageAddress == searchPageAddress) {

      /* It is.  Record that this is the index. */
      found = TRUE;

    } else {
     
      /* It isn't the index.  Increment the index and keep
         searching. */
      index++;
  
    }
  }

  /* Was the page found? */
  if (!found)

    /* It wasn't, so set index to -1 to reflect that fact. */
    index = -1;

  return index;

}
/* --------------------------------------------------------------- */



/* --------------------------------------------------------------- */
/* Given an unprotected set (dirtiedSet) and the index of a page that
   has just been given write permission (dirtyIndex), set the dirty
   field. */
void setDirtyInUnprotectedSet (UnprotectedSet* dirtiedSet,
			       int dirtyIndex) {

  /* The dirty field should not all ready be set.  If it is,
     something has gone wrong, so abort the process. */
  if (dirtiedSet->pages[dirtyIndex].dirty)
    Abort("setDirtyInUnprotectedSet: Attempted to mark an all-ready dirty page");
  
  /* Set the dirty field. */
  dirtiedSet->pages[dirtyIndex].dirty = TRUE;

}
/* --------------------------------------------------------------- */



/* --------------------------------------------------------------- */
void outputRecord (PageInfoElement* evictedPage) {
  int pageStatus;

  /* A page was selected.  Is dirty-page information being recorded? */
  if (tcc.tc->dirtyStatus == ENABLED) {
    pageStatus = ((evictedPage->dirty == FALSE) ?
		  PAGE_STATUS_READ :
		  PAGE_STATUS_WRITE);
  } else {
    pageStatus = PAGE_STATUS_NONE;
  }
  
  /* Write out a record into the trace. */
  safePrintPageNumber (tcc.tc->wbuf,
		       pageStatus,
		       PAGE_NUM (evictedPage->pageAddress));
  write_VMT(tcc.tc->fd, tcc.tc->wbuf, strlen (tcc.tc->wbuf));
  
  /* Is there an experiment attatched to this trace? */
  if (tcc.tc->exptStatus == ENABLED) {
    
    /* There is, so call on it to output its experimental information
       for this page. */
    ExptOutputData (evictedPage, tcc.tc->fd);
    
  }
}
/* --------------------------------------------------------------- */



/* --------------------------------------------------------------- */
/* This function is called by the platform-specific fault handlers.
   Once the key information is extracted from those fault handlers, the
   management is identical. */
void handleFault (caddr_t faultAddress) {
  caddr_t faultPageAddress;
  PageInfoElement* evictedPage = &(tcc.tc->evictedPage);
  int queuePosition;

  /* Calculate the address that points to the first word in this page. */
  faultPageAddress = (caddr_t)PAGE_BEG(faultAddress);

  /* Initialize the evicted page to be a null page. */
  evictedPage->pageAddress = (caddr_t)0;

  /* Are we all ready in the fault handler?  That is, did some part
     of processing a fault cause another fault? */
  if (tcc.tc->inHandler) {

    /* [Note] As a debug measure, output a marker to record that a nested
       fault occurred. */
    tcc.tc->wbuf[0] = '!';
    write_VMT(tcc.tc->fd, tcc.tc->wbuf, 1);

    /* Bail out of the process.  This is a bad sign.  Bad enough that
       we shouldn't even trust our Abort procedure. */
    exit(-1);

  } else {

    /* Record the we are in the fault handler. */
    tcc.tc->inHandler = TRUE;

  }

  /* Has dirty-bit tracing been activated? */
  if (tcc.tc->dirtyStatus == ENABLED) {

    /* Search the FIFO-queue to see if it all ready holds the page. */
    queuePosition = searchUnprotectedSet(&(tcc.tc->unprotectedSet),
					 faultPageAddress);
    
    /* Was the page found in the queue? */
    if (queuePosition == -1) {
      
      /* The page was not in the queue, so this is the first fault for
	 that page.  Set its protection to read-only. */
      if (mprotect(faultPageAddress, PAGESIZE, PROT_READ) == -1)
	
	/* The change of protection failed.  Abort the process. */
	Abort("handleFault: After first touch, couldn't set to read-only");
      
      /* Update the FIFO-queue. */
      evictedPage = addToUnprotectedSet(&(tcc.tc->unprotectedSet),
					faultPageAddress);

    } else {
      
      /* The page was all ready in the queue as a clean page.  Change
	 its protection to read/write. */
      if (mprotect(faultPageAddress,
		   PAGESIZE,
		   PROT_READ | PROT_WRITE) == -1) {
	
	/* The change of protection failed.  Abort the process. */
	Abort("handleFault: After second touch, couldn't set to read/write");
      
      }

      /* Update the FIFO-queue. */
      setDirtyInUnprotectedSet(&(tcc.tc->unprotectedSet), queuePosition);
      
    }
  } else {
    
    /* We're not tracing dirtiness.  Try to change the protection
       to read/write. */
    if (mprotect(faultPageAddress,
		 PAGESIZE,
		 PROT_READ | PROT_WRITE) == -1) {
      
      /* The change failed.  Did it fail because the page should be
	 read-only? */
      if (errno == EACCES) {

	/* This page shouldn't be read/write, so try setting it to
	   read-only. */
	if (mprotect(faultPageAddress, PAGESIZE, PROT_READ) == -1) {
	  
	  /* This failed too, abort the process. */
	  Abort("handleFault: After touch, couldn't set protection to \
                 read/write or read-only");

	}
      } else {

	/* The protection set to read/write failed for a reason we can't
	   deal with.  Abort the process. */
	Abort("handleFault: After touch, unmanagable error trying to set read/write");
	
      }
    }

    /* Update the FIFO-queue. */
    evictedPage = addToUnprotectedSet(&(tcc.tc->unprotectedSet),
				      faultPageAddress);

  }
  

  /* Was there a page selected for eviction? */
  if (evictedPage->pageAddress != (caddr_t)0) {

    /* Output a record into the trace. */
    outputRecord (evictedPage);

    /* Attempt to set the protections for the evicted page to none. */
    if (mprotect(evictedPage->pageAddress, PAGESIZE, PROT_NONE) == -1) {

      /* The change failed.  Abort the process. */
      Abort("handleFault: Could not set protections to none");

    }
  }

  /* Record that we are leaving the fault handler. */
  tcc.tc->inHandler = FALSE;

}
/* --------------------------------------------------------------- */



/* --------------------------------------------------------------- */
#ifdef __svr4__

/* Solaris specific handler */
typedef void (*PageFaultHandler) (int, siginfo_t*, void*);
void SigsegvHandler (int sig, siginfo_t* info, void* unused)
{

  /* Call the basic fault handler, which only needs some of the
     information passed to this platform-specific fault handler. */
  handleFault((caddr_t)info->si_addr);

}

#else /* not __svr4__ */
#ifdef __linux__

/* Linux specific handler */
typedef void (*PageFaultHandler) ();
void SigsegvHandler (int sig, struct sigcontext_struct scp) {

  /* Call the basic fault handler, which only needs some of the
     information passed to this platform-specific fault handler. */
  handleFault((caddr_t)scp.cr2);

}

#endif /* __linux__ */
#endif /* __svr4__ */
/* --------------------------------------------------------------- */



/* --------------------------------------------------------------- */
void createMappedSpace () {

  int mappedFile;
  caddr_t mappedSpace;
  char dummyChar;

  mappedFile = open(MAPPED_FILE_NAME, O_RDWR | O_CREAT, 0644);

  /* Was the file opening successful? */
  if (mappedFile == -1)

    /* The open failed.  Abort the process. */
    Abort("createMappedSpace: Couldn't open mapped file");

  /* Make the file large enough by seeking forward to a sufficient file
     length and then write one character. */
  lseek (mappedFile, sizeof(TraceContext), 0);
  write_VMT(mappedFile, &dummyChar, 1);

  /* mmap some space in which to store the trace context. */
  mappedSpace = mmap((caddr_t)0,
		     sizeof(TraceContext),
		     PROT_READ | PROT_WRITE,
		     MAP_PRIVATE,
		     mappedFile,
		     0);

  /* Was the mapping successful? */
#if !defined (__svr4__)
#define MAP_FAILED  (caddr_t)-1
#endif /* not __svr4__ */

  if (mappedSpace == MAP_FAILED) {

    /* The mapping failed.  Abort the process. */
    Abort("createMappedSpace: Could not mmap space for trace context");

  } else {
    
    /* The mapping worked.  Assign the trace context to that space. */
    tcc.tc = (TraceContext*)mappedSpace;
    
  }
}
/* --------------------------------------------------------------- */



/* --------------------------------------------------------------- */
void determineSettings() {
  int i;
  char* s;
  caddr_t a;
  unsigned long r;
  int queueSize;
  caddr_t currentPageAddress;
  caddr_t beginPageAddress;
  int found;

  /* Initialize the flag that indicates whether or not the fault
     handler is active. */
  tcc.tc->inHandler = FALSE;

  /* Initialize the seed used for the random number generator. */
  BSDsrand (1);

  /* Choose the default queue size, to be overridden only if specified
     by an environment variable. */
  queueSize = DEFAULT_UNPROTECTED_SET_SIZE;

  /* Get the value of the queue size environment variable. */
  s = getenv("VMT_QSIZE");

  /* Was the environment variable set? */
  if (s != (char*)0) {

    /* Convert the value of the environment variable to an integer. */
    i = atoi(s); 
    
    /* Is that value reasonable? */

    if ((i > 0) && (i < MAX_UNPROTECTED_SET_SIZE))

      /* It is, so use that value as the queue length. */
      queueSize = i;

  }

  /* Initialize the FIFO queue. */
  initUnprotectedSet(&(tcc.tc->unprotectedSet), queueSize);

  /* Grab the environment variable that indicates whether or not to use
     an attatched experiment. */
  s = getenv("VMT_EXPT");

  /* Was the environment variable set, and if so, set to true? */
  if ((s == NULL) || (strcasecmp(s,"true"))) {

    /* It wasn't set, or wasn't set to true.  Disable experiment hooks. */
    tcc.tc->exptStatus = DISABLED;

  } else {

    /* It was set to true.  Enable calls to experiment hooks, and
     call the experiment initialization hook. */
    tcc.tc->exptStatus = ENABLED;
    ExptInitSystem();

  }

  /* Grab the environment variable that stores the pathname for the
     trace output file. */
  s = getenv("VMT_OUT");

  /* Was the environment variable set at all? */
  if (s == (char*)0)
    
    /* It was not set, so use the default trace file pathname. */
    strcpy (tcc.tc->filename, DEFAULT_OUTFILE);

  else

    /* It was set, so use the pathname specified. */
    strcpy (tcc.tc->filename, s);

  /* Ensure that the file descriptor is not set to a usable value,
     as the trace file has not yet been opened. */
  tcc.tc->fd = -1;
  
  /* Get the value of the environment variable VMT_DIRTY. */
  s = getenv("VMT_DIRTY");
  
  /* Is VMT_DIRTY set to "true"? */
  if ((s == NULL) || (strcasecmp(s,"true")))
    
    /* It isn't, so disable dirty-bit tracing. */
    tcc.tc->dirtyStatus = DISABLED;
  
  else
    
    /* It is, so enable dirty-bit tracing. */
    tcc.tc->dirtyStatus = ENABLED;
  
  /* Get the value of the enviroment variable VMT_POLICY. */
  s = getenv("VMT_POLICY");
  
  /* Should the policy be set to the default, FIFO? */
  if ((s == NULL) ||
      (strcasecmp (s,"FIFO") == 0) ||
      (strcasecmp (s, "RANDOM") != 0)) {

    /* Set the replacement policy for unprotected pages to FIFO. */
    tcc.tc->replacement = FIFO;

  } else {

    /* Set the policy to RANDOM, the other replacement policy so
       far. */
    tcc.tc->replacement = RANDOM;

  }

  /*
   ** Allocate an extra page so that heap and static data don't
   ** share a page, round the allocation boundary to a pagesize,
   ** and set the internal markers at the boundary.
   */
  a = __real_sbrk(PAGESIZE);

  /* Did the allocation succeed? */
  if (a == (caddr_t)-1)

    /* It failed, so abort the process. */
    Abort("VMTrace: Region disjoining sbrk failed");

  if ((r= PAGE_OFF(a)) != 0)
    if ((a= __real_sbrk(PAGESIZE-r)) == (caddr_t)-1)
      Abort("VMTrace: Page rounding sbrk failed");
  
  /*
   ** abrk is the actual break; ubrk is the view given to the
   ** program calling sbrk(). ubrk rounded up to the next page
   ** boundary is abrk. Reason for having both: when someone
   ** asks for a fractional page from sbrk(), to protect it
   ** one has to allocate the full page.
   */
  
  tcc.tc->start_of_heap =
    tcc.tc->abrk =
    tcc.tc->ubrk =
    (caddr_t)((unsigned long)a+(PAGESIZE-r));
  
  /* Get the value of the environment variable VMT_STATIC. */
  s = getenv("VMT_STATIC");
  
  /* Is VMT_STATIC set to "true"? */
  if ((s == NULL) || (strcasecmp(s,"true")))
    
    /* It isn't, so disable static data segment tracing. */
    tcc.tc->staticStatus = DISABLED;
  
  else {
    
    /* It is, so enable static data segment tracing. */
    tcc.tc->staticStatus = ENABLED;
    
    /* We need to find the data segment.  It ends at &_end, and
       begins either at &_etext (Linux), or at some invalid page
       (Solaris).  Use mprotect going backwards from &_end until
       one of those two boundaries is encountered, and that
       will be the beginning of the data segment. */
    tcc.tc->dataSegmentEnd = (caddr_t) &_end;

    /* Assume that the data segment will begin at the page
       immediately following &_etext unless we discover
       otherwise.  Note that we don't protect the page on
       which &_etext lies, as it we don't know what will
       happen if we protect program text. */
    beginPageAddress = (caddr_t) PAGE_NXT (&_etext);

    /* Set the currrent page address to be the beginning of the
       page on which &_end lies. */
    currentPageAddress = (caddr_t) PAGE_BEG (&_end);

    /* Proceed backwards through memory one page at a time,
       calling mprotect (but leaving all permissions in tact),
       until either beginPageAddress or an invalid page is
       encountered. */
    found = FALSE;
    while ((!found) && (currentPageAddress >= beginPageAddress)) {

      /* Is the current page valid? */
      if (mprotect (currentPageAddress,
		    PAGESIZE,
		    PROT_READ | PROT_WRITE | PROT_EXEC) == -1) {

	/* It is not, so we've found the beginning of the data
	   segment. */
	found = TRUE;

      } else {

	/* It is, so try to protect the previous page. */
	currentPageAddress -= PAGESIZE;

      }
    }

    /* Was an invalid page found? */
    if (found) {

      /* There was, so the beginning of the data segment is one
	 page beyond the last page we tried to protect. */
      tcc.tc->dataSegmentBegin = currentPageAddress + PAGESIZE;

    } else {

      /* There wasn't.  The data segment begins right after the
	 end of the text segment. */
      tcc.tc->dataSegmentBegin = beginPageAddress;

    }
  }
}
/* --------------------------------------------------------------- */



/* --------------------------------------------------------------- */
void VMTStart () {

  TraceContext** tempTCPtr;
  struct sigaction action;

  /* Record that tracing is enabled. */
  tcc.status = ENABLED;

  /* Set up the signal handler to call our function for SIGSEGV */ 
#ifdef __svr4__
  action.sa_sigaction = (PageFaultHandler) SigsegvHandler;
  action.sa_flags = SA_SIGINFO | SA_RESTART;
#else
#ifdef __linux__
  action.sa_handler = (PageFaultHandler) SigsegvHandler;
  action.sa_flags = SA_RESTART;
#endif /* __linux__ */
#endif /* __svr4__ */
  sigfillset(&(action.sa_mask));
  if (sigaction (SIGSEGV, &action, NULL) == -1)
    Abort("VMTrace: Error installing signal handler");
  
  /* Attempt to open the trace output file. */
  tcc.tc->fd = open(tcc.tc->filename, O_CREAT | O_WRONLY | O_TRUNC, 0644);
  
  /* Did the open succeed? */
  if (tcc.tc->fd == -1)
    
    /* The open failed.  Abort the proceess. */
    Abort("VMTrace: Cannot open trace output file");
  
  /* Is static data segment tracing active?  And is there a static
     data segment to protect? */
  if ((tcc.tc->staticStatus == ENABLED) &&
      (tcc.tc->dataSegmentEnd > tcc.tc->dataSegmentBegin)) {
    
    /* There is.  First, grab the address of the pointer into the
       trace context. */
    tempTCPtr = &tcc.tc;
    
#if defined (__svr4__) /* Solaris */
    
    /* Provided that _siguhandler doesn't reside on the first page
       of the segment, protect from the beginning of the static
       data segment up to the page holding _siguhandler. */
    if ((caddr_t) PAGE_BEG (&_siguhandler) != tcc.tc->dataSegmentBegin) {
      if (mprotect (tcc.tc->dataSegmentBegin,
		    (size_t) (PAGE_BEG (&_siguhandler) -
			      (unsigned long) tcc.tc->dataSegmentBegin),
		    PROT_NONE) == -1) {
	
	/* The protection failed, so abort the process. */
	perror ("VMTrace: Error protecting data segment part 1");
	
      }
    }
    
    /* Provided that _siguhandler doesn't reside on the last page
       of the segment, protect from the page after the page holding
       _siguhandler to the end of the data segment. */
    if ((caddr_t) PAGE_BEG (&_siguhandler) != tcc.tc->dataSegmentEnd) {
      if (mprotect ((caddr_t) PAGE_NXT (&_siguhandler),
		    (size_t) (tcc.tc->dataSegmentEnd -
			      (unsigned long) PAGE_NXT (&_siguhandler)),
		    PROT_NONE) == -1) {
	
	/* The protection failed, so abort the process. */
	perror ("VMTrace: Error protecting data segment part 2");
	
      }
    }
    
#else
    
    /* Protect the entire static data segment. */
    if (mprotect (tcc.tc->dataSegmentBegin, tcc.tc->dataSegmentEnd - 
		  tcc.tc->dataSegmentBegin, PROT_NONE) == -1) {
      
      /* The protection failed, so abort the process. */
      Abort("VMTrace: Error protecting data segment");
      
    }
    
#endif /* __svr4__ */
    
    /* Unprotect the page on which the trace context pointer lies.
       Note that the trace context container structure guarantees
       that no other data is store on the same page as that pointer. */
    if (mprotect ((caddr_t) PAGE_BEG (tempTCPtr),
		  PAGESIZE,
		  PROT_READ | PROT_WRITE) == -1) {
      
      /* The unprotection failed, so abort the process. */
      Abort("VMTrace: Error unprotecting trace context pointer page");
      
    }
  }
  
  /* Is the actual break at the start of the heap? */
  if (tcc.tc->abrk != tcc.tc->start_of_heap) {
    
    /* It isn't, so protect everything from the start of the heap
       through the actual break. */
    if (mprotect (tcc.tc->start_of_heap, tcc.tc->abrk - 
		  tcc.tc->start_of_heap,
		  PROT_NONE) == -1)
      
      /* The protection failed.  Abort the proceess. */
      Abort ("VMTrace: Error protecting existing heap region");
    
  }
}
/* --------------------------------------------------------------- */



/* --------------------------------------------------------------- */
/*
** Intialize the tracing module
*/
void InitVMTrace() {
  char* s;
  
  /* Grab the environment variable that indicates whether or not
     tracing should be active. */
  s = getenv("VMT_FLAG");

  /* Was the environment variable set, and if so, set to true? */
  if (s == (char*)0 || strcasecmp(s, "true")) {

    /* It was not set, so do not activate tracing. */
    tcc.status = DISABLED;
    
  } else {

    /* It was set.  Create a mapped space in which VMTrace can store
       its state. */
    createMappedSpace ();

    /* Extract the desired user settings from environment
       and set up for tracing. */
    determineSettings ();

    /* Begin tracing. */
    VMTStart ();

  }
}
/* --------------------------------------------------------------- */



/* --------------------------------------------------------------- */
void VMTFlush () {

  int counter;
  PageInfoElement* evictedPage;

  /* Are there any pages left in the FIFO queue?  Insert null pages to
     flush out any remaining pages being stored. */
  for (counter = 0; counter < tcc.tc->unprotectedSet.size; counter++) {

    /* Insert a null page into the FIFO queue. */
    evictedPage = addToUnprotectedSet (&(tcc.tc->unprotectedSet),
				       (caddr_t)0);

    /* Was a page evicted? */
    if (evictedPage->pageAddress != (caddr_t)0) {

      /* Output a record into the trace. */
      outputRecord (evictedPage);

    }
  }
}
/* --------------------------------------------------------------- */



/* --------------------------------------------------------------- */
void VMTStop () {

  /* Disable the signal handler. */
  if (sigaction(SIGSEGV, NULL, NULL) == -1) {
    Abort("VMTrace: Error removing signal handler");
  }

  /* Remove protections from the heap. */
  if (mprotect (tcc.tc->start_of_heap,
		tcc.tc->abrk - tcc.tc->start_of_heap, 
		PROT_READ | PROT_WRITE) == -1)

    /* The change in protection failed.  Abort the process. */
    Abort ("VMTrace: Error unprotecting existing heap region");

  /* Was any of the static data segment protected? */
  if (tcc.tc->dataSegmentBegin < tcc.tc->dataSegmentEnd) {

    /* There was, so remove protections from the static data segment. */
    if (mprotect (tcc.tc->dataSegmentBegin, 
		  tcc.tc->dataSegmentEnd - tcc.tc->dataSegmentBegin,
		  PROT_READ | PROT_WRITE | PROT_EXEC) == -1)
      
      /* The change in protection failed.  Abort the process. */
      Abort("VMTrace: Error unprotecting static data segment");
    
  }

  /* Close the output trace file. */
  close (tcc.tc->fd);

  /* Record that tracing has been disabled. */
  tcc.status = DISABLED;

  /* Unmap the mapped space. */
  if (munmap((caddr_t)tcc.tc, sizeof(TraceContext)) == -1) {

    /* The unmapping failed.  Abort the process. */
    Abort("VMTrace: Error unmapping trace context space");

  }

  /* Set the trace context point to NULL, so that nobody mistakenly
     tries to use it. */
  tcc.tc = NULL;

}
/* --------------------------------------------------------------- */



/* --------------------------------------------------------------- */
/*
 ** Shut down the tracing module
 */
void ShutdownVMTrace() {

  /* Was VMTrace ever activated? */
  if (tcc.tc != NULL) {

    /* Flush out records for all of the pages still sitting in the
       unprotected set. */
    VMTFlush();

    /* It was activated.  Is there an attatched experiment? */
    if (tcc.tc->exptStatus == ENABLED)

      /* There is, so call the hook to shut it down. */
      ExptShutdownSystem(tcc.tc->fd);
  
    /* Clean up VMTrace itself. */
    VMTStop();

  }  
}
/* --------------------------------------------------------------- */
